/*
   3D Invaders

   game.cc

   (c)2002 Paul Boese
*/

#include "common.h"
#include "game.h"
#include <math.h>

//extern "C" {
//long rand();
//}

// Preliminary movement
enum { MV_HOME, MV_RT, MV_DN, MV_LT, MV_UP };

horde_t Game::horde[IROWS][ICOLS];
block_t Game::block[BARRIERS];

int Game::move;
int Game::xx;
int Game::yy;

int Game::refcnt = 0;

Game::Game() {
	mode = GSTART;
	score = 0;
	hi_score = 0;
	level = 0;
	shots = 0;
	ticks = 0;
	refcnt++;
	move = MV_HOME;
	xx = 0;
	yy = 0;
}

void Game::doInitLevel(Invader ** invaders, Barrier ** barriers,
			Player * player) {
	float layer[9][2] = {
		{ -5.0f,  5.0f }, {  0.0f,  5.0f }, {  5.0f,  5.0f },
		{ -5.0f,  0.0f }, {  0.0f,  0.0f }, {  5.0f,  0.0f },
		{ -5.0f, -5.0f }, {  0.0f, -5.0f }, {  5.0f, -5.0f }
	};
	int i, j, k;
	float iz=-30.0f;  // Inital Z position for first row for invaders
	float bz =-10.0f; // Barrier Z position
	// Player Z position
	point_t ppos;
	ppos.x = ppos.y = 0.0f; ppos.z = 5.0f;

	ticks = 0;

	move = MV_HOME;
	xx = 0;
	yy = 0;
	
	for (i=0, k=0; i<IROWS; i++) {
		for (j=0; j<ICOLS; j++, k++) {
			horde[i][j].active = -1;
			horde[i][j].pos.x = layer[j][0];
			horde[i][j].pos.y = layer[j][1];
			horde[i][j].pos.z = iz;
			horde[i][j].row = i;
			horde[i][j].col = j;
			horde[i][j].bomber = !i ? CAN_BOMB : BOMB_NOT;
			invaders[k]->Reset();
		}
		iz += -10.0f; // Z distance between invaders
	}

	for (i=0; i<BARRIERS; i++) {
		block[i].active = -1;
		block[i].pos.x = layer[i][0];
		block[i].pos.y = layer[i][1];
		block[i].pos.z = bz;
		barriers[i]->Reset(&block[i].pos);

	}

	player->Reset(&ppos);
}

#define FABS(n) ((((n))<0.0f)? -((n)) : ((n)))

int Game::hit(point_t *v1, point_t *v2, float size) {
	float x, y, z, d;
	x = (v1->x) - (v2->x);
	y = (v1->y) - (v2->y);
	z = (v1->z) - (v2->z);
	d = fsqrt( (x*x) + (y*y) + (z*z) );
	return (d <= size)? -1 : 0;
}

// Promote the next bomber in the column to CAN_BOMB status (if possible)
void Game::promote_bomber(int row, int col) {
	int i=1;
	
	horde[row][col].bomber = BOMB_NOT; // This one's gone so demote
	
	do {
		if ( (row+i) < IROWS ) {
			if (horde[row+i][col].active){
				horde[row+i][col].bomber = CAN_BOMB;
				i=IROWS; // force exit from loop
			}
		}
		i++;
	} while ((row+i) < IROWS);
}

#define BLIST 9

void Game::doInvaders(Invader ** invaders, Shots ** shots, Bombs ** bombs) {
	int i, j, k, m;
	point_t *vi, *vs;
	float x, y, z;
	int bs[BLIST][2];  // eligable bombers list
	static int cb=0;   // current bomb

	x = y = z = 0.0f;

	ticks++;
	// Check for hits between active invaders and shots.
	// Falling invaders absorb shots.
	for (i=0, k=0; i<IROWS; i++) {
		for (j=0; j<ICOLS; j++, k++) {
			for (m=0; m<MAX_SHOTS; m++) {
				if (shots[m]->Active()) {
					vs = shots[m]->Pos();
					vi = invaders[k]->Pos();
					if (hit(vi, vs, 1.414f)) {
						shots[m]->Hit();
						invaders[k]->Hit();
						horde[i][j].active = 0;
						promote_bomber(i,j);
					}
				}
			}
		}
	}
	
	// Draw the horde of invaders
	for (i=0, k=0; i<IROWS; i++) {
		for (j=0; j<ICOLS; j++, k++) {
			if(invaders[k]->Active())
				invaders[k]->Draw(&horde[i][j].pos);
		}
	}

	// Move Invaders
	switch (move) {
	case MV_HOME:
		x = -0.04f;
		y = -0.04f;
		z = 0.0f;
		if (xx++ >= 100) {
			xx = 0;
			move = MV_RT;
		}
		break;
	case MV_RT:
		x = 0.04f;
		y = 0.0f;
		z = 0.0f;
		if (xx++ >= 200) {
			xx = 0;
			move = MV_DN;
		}
		break;
	case MV_DN:
		x = 0.0f;
		y = 0.04f;
		z = 0.0f;
		if (xx++ >= 200) {
			xx = 0;
			move = MV_LT;
		}
		break;
	case MV_LT:
		x = -0.04f;
		y = 0.0f;
		z = 0.0f;
		if (xx++ >= 200) {
			xx = 0;
			move = MV_UP;
		}
		break;
	case MV_UP:
		x = 0.0f;
		y = -0.04f;
		z = 0.01;
		if (xx++ >= 200) {
			xx = 0;
			move = MV_RT;
		}
		break;
	}

	// Update the hordes postition
	for (i=0, k=0; i<IROWS; i++) {
		for (j=0; j<ICOLS; j++, k++) {
			if(horde[i][j].active) {
				horde[i][j].pos.x += x;
				horde[i][j].pos.y += y;
				horde[i][j].pos.z += z;
			}
		}
	}

	// Drop a bomb every 1 second
	// Select a bomber ramdomly from the eligiable bombers
	if (!(ticks % (1*60))) {
		for (i=0, k=0; i<IROWS; i++) {
			for (j=0; j<ICOLS; j++) {
				if (horde[i][j].active &&
				    horde[i][j].bomber &&
				    k < BLIST ) {
					bs[k][0]=i;
					bs[k][1]=j;
					k++;
				}
			}
		}
		k = rand() % (k);
		if (k >= 0 && k < BLIST)
			if (horde[bs[k][0]][bs[k][1]].active)
				bombs[cb]->Drop(&horde[bs[k][0]][bs[k][1]].pos);
		cb = ++cb % (MAX_BOMBS-1);
	}
	
}

void Game::doBarrier(Barrier ** barriers, Invader ** invaders,
		     Shots ** shots, Bombs ** bombs) {
	int i, j, k, m;
	point_t *vs, *vb, *vi;

	// Check for hits between shots and barriers
	// Player can blow holes in the barrier!
	// Not fully implemented yet!
	for (i=0; i<BARRIERS; i++) {
		for (j=0; j<MAX_SHOTS; j++) {
			if (shots[j]->Active()) {
				vs = shots[j]->Pos();
				vb = barriers[i]->Pos();
				if (hit(vb, vs, 2.0f)) {
					shots[j]->Hit();
					barriers[i]->Hit();
					block[i].active = 0;
				}
			}
		}
	}

	// Check for hits between bombs and barriers
	for (i=0; i<BARRIERS; i++) {
		for (j=0; j<MAX_BOMBS; j++) {
			if (bombs[j]->Active()) {
				vs = bombs[j]->Pos();
				vb = barriers[i]->Pos();
				if (hit(vb, vs, 2.0f)) {
					bombs[j]->Hit();
					barriers[i]->Hit();
					block[i].active = 0;
				}
			}
		}
	}
	
	// Check for collisions between invaders and barriers
	// Falling Invaders can weaken the barrier!
	// When invaders meet the barrier they are destoryed!
	for (i=0; i<BARRIERS; i++) {
		if (barriers[i]->Active()) {
			for (j=0, m=0; j<(IROWS); j++) {
				for (k=0; k<ICOLS; k++, m++) {
					vi = invaders[m]->Pos();
					vb = barriers[i]->Pos();
					if (hit(vb, vi, 2.75f)) {
						invaders[m]->Poof();
						barriers[i]->Hit();
						promote_bomber(j,k);
						block[i].active = 0;
					}
				}
			}
		}
	}

	// Draw the barriers
	for (i=0; i<BARRIERS; i++) {
		if(barriers[i]->Active())
			barriers[i]->Draw(&block[i].pos);
	}
}

void Game::doPlayer(Player * player, Invader ** invaders, Bombs ** bombs) {
	int i, j, k, m;
	point_t *vp, *vb;
	player->Draw();
	
	// Check for collision between player and invaders
	for (j=0, m=0; j<(IROWS); j++) {
		for (k=0; k<ICOLS; k++, m++) {
			vb = invaders[m]->Pos();
			vp = player->Pos();
			if (hit(vb, vp, 2.0f)) {
				invaders[m]->Poof();
				horde[i][j].active = 0;
				player->Hit();
			}
		}
	}

	// Check for collisions between player and bombs
	for (i=0; i<MAX_BOMBS; i++) {
		if (bombs[i]->Active()) {
			vb = bombs[i]->Pos();
			vp = player->Pos();
			if (hit(vb, vp, 1.414f)) {
				bombs[i]->Hit();
				player->Hit();
			}
		}
	}
	player->Draw();
}

